# -*- coding: utf-8 -*-

import json

from ..utils.encoding import encode_tx_hash
from ..status.commit_status import CommitStatus


class TxResult:
    def __init__(self, tx=None, result=None):
        self.tx = tx
        self.__result = result

        if result is not None:
            self.tx_id = encode_tx_hash(self.__result.hash)
            self.status = CommitStatus(self.__result.error)
            self.detail = self.__result.detail

        self.contract_address = None

    def __str__(self):
        result = {
            'tx_id': self.tx_id,
            'status': self.status.name,
            'detail': self.detail,
            'contract_address': self.contract_address
        }
        return json.dumps(result, indent=2)



# -*- coding: utf-8 -*-

import base58


class BlockHash:
    def __init__(self, bh):
        if isinstance(bh, str):
            bh = BlockHash.decode_block_hash(bh)
        self.__block_hash = bh

    @property
    def value(self):
        return self.__block_hash

    def __bytes__(self):
        return self.__block_hash

    def __str__(self):
        return BlockHash.encode_block_hash(self.__block_hash)

    @staticmethod
    def encode_block_hash(b):
        return base58.b58encode(b).decode('utf-8')

    @staticmethod
    def decode_block_hash(s):
        return base58.b58decode(s)



# -*- coding: utf-8 -*-

"""Common utility module for converting types."""

import json
import toml
import socket

from ..obj.aergo_conf import AergoConfig
from ..grpc import blockchain_pb2


def convert_toml_to_aergo_conf(v):
    aergo_conf = AergoConfig()

    conf = toml.loads(v)
    for k, v in conf.items():
        if isinstance(v, dict):
            for k2, v2 in v.items():
                aergo_conf.add_conf(k2, v2, k)
        else:
            aergo_conf.add_conf(k, v)

    return aergo_conf


def convert_aergo_conf_to_toml(aergo_conf):
    return toml.dumps(aergo_conf.conf)


def convert_tx_to_grpc_tx(tx):
    grpc_tx = blockchain_pb2.Tx()
    grpc_tx.hash = bytes(tx.tx_hash)
    if tx.nonce is not None:
        grpc_tx.body.nonce = tx.nonce
    if tx.from_address is not None:
        grpc_tx.body.account = bytes(tx.from_address)
    if tx.to_address is not None:
        grpc_tx.body.recipient = bytes(tx.to_address)
    if tx.amount is not None:
        grpc_tx.body.amount = bytes(tx.amount)
    if tx.payload is not None:
        grpc_tx.body.payload = tx.payload
    grpc_tx.body.limit = tx.fee_limit
    grpc_tx.body.price = bytes(tx.fee_price)
    grpc_tx.body.type = tx.tx_type.value
    if tx.sign is not None:
        grpc_tx.body.sign = tx.sign
    return grpc_tx


def tx_to_grpc_tx(v):
    return convert_tx_to_grpc_tx(v)


def convert_tx_to_json(tx):
    if tx is None:
        return None

    return tx.json()


def tx_to_json(v):
    return convert_tx_to_json(v)


def tx_to_json(v):
    return convert_tx_to_json(v)


def convert_tx_to_formatted_json(tx):
    if tx is None:
        return None
    return json.dumps(convert_tx_to_json(tx), indent=2)


def tx_to_formatted_json(v):
    return convert_tx_to_formatted_json(v)


def convert_bytes_to_int_str(v):
    return ''.join('{:d} '.format(x) for x in v)


def bytes_to_int_str(v):
    return convert_bytes_to_int_str(v)


def convert_bytes_to_hex_str(v):
    return ''.join('0x{:02x} '.format(x) for x in v)


def convert_ip_bytes_to_str(v):
    if isinstance(v, str):
        return v

    l = len(v)

    # IPv4
    if 4 == l:
        return socket.inet_ntoa(v)
    elif 16 == l and all(v2 == 0 for v2 in list(v[:10])) and 255 == v[10] and 255 == v[11]:
        return socket.inet_ntoa(v[12:16])

    # IPv6
    return socket.inet_ntop(socket.AF_INET6, v)


""" Deprecated
def convert_luajson_to_json(v):
    v = v.decode('utf-8').replace('\\', '')
    v = v[1:len(v)-1]
    return json.loads(v)
"""


def convert_bigint_to_bytes(number):
    q, r = divmod(len(bin(number))-2, 8)
    bytes_to_fit_number = q if r == 0 else q + 1
    return number.to_bytes(bytes_to_fit_number, 'big')


def bigint_to_bytes(v):
    return convert_bigint_to_bytes(v)



import pytest

import aergo.herapy as herapy


DEFAULT_AERGO_CONFIG = """datadir = "${AERGO_HOME}/data"
dbtype = "badgerdb"
enableprofile = false
profileport = 6060
enablerest = false
enabletestmode = false
personal = true

[rpc]
netserviceaddr = "127.0.0.1"
netserviceport = 7845
netservicetrace = false
nstls = false
nskey = ""

[rest]
restport = 8080

[p2p]
netprotocoladdr = ""
netprotocolport = 7846
npbindaddr = ""
npbindport = -1
nptls = false
npcert = ""
npkey = ""
npaddpeers = []
nphiddenpeers = []
npdiscoverpeers = true
npmaxpeers = 100
nppeerpool = 100
npexposeself = true
npusepolaris = true
npaddpolarises = []

[polaris]
allowprivate = false
genesisfile = ""

[blockchain]
maxblocksize = 1048576
coinbaseaccount = ""
maxanchorcount = 20
usefastsyncer = false

[mempool]
showmetrics = false
enablefadeout = false
fadeoutperiod = 12
dumpfilepath = "${AERGO_HOME}/mempool.dump"

[consensus]
enablebp = false
blockinterval = 1

[monitor]
protocol = ""
endpoint = ""

[account]
unlocktimeout = 60
"""


def test_default():
    aergo_conf = herapy.AergoConfig()
    generate_toml = herapy.utils.convert_aergo_conf_to_toml(aergo_conf)
    assert DEFAULT_AERGO_CONFIG == generate_toml
    with pytest.raises(KeyError):
        aergo_conf.rpc_nscert
    with pytest.raises(KeyError):
        aergo_conf.rpc_nsallowcors
    with pytest.raises(KeyError):
        aergo_conf.blockchain_verifiercount
    with pytest.raises(KeyError):
        aergo_conf.mempool_verifiers


TEST_AERGO_CONFIG = """
# aergo TOML Configuration File (https://github.com/toml-lang/toml)
# base configurations
datadir = "/Users/yp/work/blocko/go/src/github.com/aergoio/aergo/bin/alone/data"
enableprofile = false
profileport = 6060
enablerest = false
enabletestmode = true

[rpc]
netserviceaddr = "127.0.0.1"
netserviceport = 7845
nstls = false
nscert = ""
nskey = ""
nsallowcors = false

[rest]
restport = "8080"

[p2p]
netprotocoladdr = "127.0.0.1"
netprotocolport = 7846
nptls = false
npcert = ""
npkey = ""
npaddpeers = [
]
npmaxpeers = "100"
nppeerpool = "100"

[blockchain]
# blockchain configurations
maxblocksize = 1048576

[mempool]
showmetrics = false
dumpfilepath = "/Users/yp/work/blocko/go/src/github.com/aergoio/aergo/bin/alone/mempool.dump"

[consensus]
enablebp = true
blockinterval = 3
"""


def test_success():
    aergo_conf = herapy.utils.convert_toml_to_aergo_conf(TEST_AERGO_CONFIG)
    with pytest.raises(KeyError):
        aergo_conf.add_conf("test", "test")
    with pytest.raises(KeyError):
        aergo_conf.add_conf("test", "test", "test")
    with pytest.raises(KeyError):
        aergo_conf.add_conf("test", "test", "rpc")
    # check base config
    assert aergo_conf.datadir == "/Users/yp/work/blocko/go/src/github.com/aergoio/aergo/bin/alone/data"
    assert aergo_conf.dbtype == herapy.AERGO_DEFAULT_CONF['dbtype']
    assert aergo_conf.enableprofile is False
    assert aergo_conf.profileport == 6060
    assert aergo_conf.enablerest is False
    assert aergo_conf.enabletestmode is True
    assert aergo_conf.personal is herapy.AERGO_DEFAULT_CONF['personal']
    # check rcp config
    assert aergo_conf.rpc_netserviceaddr == "127.0.0.1"
    assert aergo_conf.rpc_netserviceport == 7845
    assert aergo_conf.rpc_netservicetrace is False
    assert aergo_conf.rpc_nstls is False
    with pytest.raises(KeyError):
        herapy.AERGO_DEFAULT_CONF['rpc']['nscert']
    assert aergo_conf.rpc_nscert == ""
    assert aergo_conf.rpc_nskey == ""
    with pytest.raises(KeyError):
        herapy.AERGO_DEFAULT_CONF['rpc']['nsallowcors']
    assert aergo_conf.rpc_nsallowcors is False
    # check rest config
    assert aergo_conf.rest_restport == 8080
    # check p2p config
    assert aergo_conf.p2p_netprotocoladdr != herapy.AERGO_DEFAULT_CONF['p2p']['netprotocoladdr']
    assert aergo_conf.p2p_netprotocoladdr == "127.0.0.1"
    assert aergo_conf.p2p_netprotocolport == 7846
    assert aergo_conf.p2p_npbindaddr == herapy.AERGO_DEFAULT_CONF['p2p']['npbindaddr']
    assert aergo_conf.p2p_npbindport == herapy.AERGO_DEFAULT_CONF['p2p']['npbindport']
    assert aergo_conf.p2p_nptls is False
    assert aergo_conf.p2p_npcert == ""
    assert aergo_conf.p2p_npkey == ""
    assert isinstance(aergo_conf.p2p_npaddpeers, list)
    assert 0 == len(aergo_conf.p2p_npaddpeers)
    assert aergo_conf.p2p_npmaxpeers == 100
    assert aergo_conf.p2p_nppeerpool == 100
    # check blockchain config
    assert aergo_conf.blockchain_maxblocksize == 1048576
    assert aergo_conf.blockchain_coinbaseaccount == herapy.AERGO_DEFAULT_CONF['blockchain']['coinbaseaccount']
    assert aergo_conf.blockchain_maxanchorcount == herapy.AERGO_DEFAULT_CONF['blockchain']['maxanchorcount']
    assert aergo_conf.blockchain_usefastsyncer is herapy.AERGO_DEFAULT_CONF['blockchain']['usefastsyncer']
    # check mempool config
    assert aergo_conf.mempool_showmetrics is False
    with pytest.raises(KeyError):
        herapy.AERGO_DEFAULT_CONF['mempool']['verifiers']
    with pytest.raises(KeyError):
        aergo_conf.mempool_verifiers
    assert aergo_conf.mempool_dumpfilepath == "/Users/yp/work/blocko/go/src/github.com/aergoio/aergo/bin/alone/mempool.dump"
    # check consensus config
    assert aergo_conf.consensus_enablebp is True
    assert aergo_conf.consensus_blockinterval == 3
    # check monitor config
    assert aergo_conf.monitor_protocol == herapy.AERGO_DEFAULT_CONF['monitor']['protocol']
    assert aergo_conf.monitor_endpoint == herapy.AERGO_DEFAULT_CONF['monitor']['endpoint']



# -*- coding: utf-8 -*-

"""Enumeration of Smart contract Status."""

import enum


@enum.unique
class SmartcontractStatus(enum.Enum):
    CREATED = "CREATED"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"



import sys
import traceback
import json

import aergo.herapy as herapy
from aergo.herapy.utils.converter import convert_bytes_to_hex_str, \
                                         convert_bytes_to_int_str


def run():
    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Get Blockchain Status -----------")
        best_block_hash, best_block_height = aergo.get_blockchain_status()
        print("Best Block Hash      = {}".format(best_block_hash))
        print("str(Best Block Hash)     = {}".format(str(best_block_hash)))
        print("bytes(Best Block Hash)   = {}".format(bytes(best_block_hash)))
        print("int(Best Block Hash)     = {}"
              .format(convert_bytes_to_int_str(bytes(best_block_hash))))
        print("hex(Best Block Hash)     = {}"
              .format(convert_bytes_to_hex_str(bytes(best_block_hash))))
        print("Best Block Height    = {}".format(best_block_height))

        print("------ Get Block Status -----------")
        block = aergo.get_block(best_block_hash)
        print(json.dumps(block.json(), indent=2))
        print("Block Hash = ", block.hash)
        print("  Height = ", block.height)
        print("  Timestamp = ", block.timestamp)
        print("  Blocks Root Hash = ", block.blocks_root_hash)
        print("  Txs Root Hash = ", block.txs_root_hash)
        print("  Confirm = ", block.confirms)
        print("  Pub Key = ", block.public_key)
        print("  Sign = ", block.sign)
        print("  previous block hash =\n{}".format(block.prev))
        print("  Txs = ", block.tx_list)

        block = aergo.get_block(block_height=best_block_height)
        print(block)
        print("Block Hash = ", block.hash)
        print("  Height = ", block.height)
        print("  Timestamp = ", block.timestamp)
        print("  Blocks Root Hash = ", block.blocks_root_hash)
        print("  Txs Root Hash = ", block.txs_root_hash)
        print("  Confirm = ", block.confirms)
        print("  Pub Key = ", block.public_key)
        print("  Sign = ", block.sign)
        print("  previous block hash =\n{}".format(block.prev))
        print("  Txs = ", block.tx_list)

        print("------ Get Peers -----------")
        # definition of peer needs to describe.
        peers = aergo.get_peers()
        for p in peers:
            print(p)

        print("------ Get Node State -----------")
        node_state = aergo.get_node_state()
        print(node_state)
        node_state_fmt_txt = json.dumps(node_state, indent=2, sort_keys=True)
        print(node_state_fmt_txt)

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception:
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-

import hashlib

from google.protobuf.json_format import MessageToJson
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from .obj import address as addr
from .obj import aer
from .obj import private_key as pk

from .utils.encoding import decode_address, decode_root
from .utils import merkle_proof as mp

from .errors.general_exception import GeneralException


class Account:
    """ Account can be a user account with private and public key,
    or a contract account.
    """
    def __init__(self, private_key=None, empty=False):
        if empty:
            self.__private_key = None
            self.__address = None
            self.__state = None
            self.__state_proof = None
            return

        self.__private_key = pk.PrivateKey(private_key)
        self.__address = addr.Address(self.__private_key.public_key)
        self.__state = None
        self.__state_proof = None

    def __str__(self):
        if self.__state_proof:
            return MessageToJson(self.__state_proof)
        elif self.__state:
            return self.state
        return super().__str__()

    def sign_msg_hash(self, msg_hash):
        return self.__private_key.sign_msg(msg_hash)

    def verify_sign(self, msg_hash, sign):
        return self.__private_key.verify_sign(msg_hash, sign)

    @property
    def private_key(self):
        return self.__private_key

    @property
    def public_key(self):
        if self.__private_key is None:
            return None

        return self.__private_key.public_key

    @property
    def address(self):
        return self.__address

    @address.setter
    def address(self, v):
        if self.__address is not None:
            raise ValueError('not empty account')

        if isinstance(v, str):
            v = decode_address(v)
        self.__address = v

    @property
    def state(self):
        if self.__state is None:
            return None
        return MessageToJson(self.__state)

    @state.setter
    def state(self, v):
        self.__state = v

    @property
    def state_proof(self):
        return self.__state_proof

    @state_proof.setter
    def state_proof(self, v):
        self.__state_proof = v

    @property
    def nonce(self):
        if self.__state is None:
            return -1
        return self.__state.nonce

    @nonce.setter
    def nonce(self, v):
        if self.__state.nonce >= v:
            return
        self.__state.nonce = v

    @property
    def balance(self):
        if self.__state is None:
            return aer.Aer()
        return aer.Aer(int.from_bytes(self.__state.balance, 'big'))

    @property
    def code_hash(self):
        if self.__state is None:
            return None
        return self.__state.codeHash

    @property
    def storage_root(self):
        if self.__state is None:
            return None
        return self.__state.storageRoot

    @property
    def sql_recovery_point(self):
        if self.__state is None:
            return None
        return self.__state.sqlRecoveryPoint

    @staticmethod
    def encrypt_account(account, password):
        """
        https://cryptography.io/en/latest/hazmat/primitives/aead/
        :param account: account to export
        :return: encrypted account data (bytes)
        """
        if isinstance(password, str):
            password = bytes(password, encoding='utf-8')

        m = hashlib.sha256()
        m.update(password)
        hash_pw = m.digest()

        m = hashlib.sha256()
        m.update(password)
        m.update(hash_pw)
        enc_key = m.digest()

        nonce = hash_pw[4:16]
        aesgcm = AESGCM(enc_key)
        return aesgcm.encrypt(nonce=nonce,
                              data=bytes(account.private_key),
                              associated_data=b'')

    @staticmethod
    def decrypt_account(encrypted_bytes, password):
        """
        https://cryptography.io/en/latest/hazmat/primitives/aead/
        :param encrypted_bytes: encrypted data (bytes) of account
        :param password: to decrypt the exported bytes
        :return: account instance
        """
        if isinstance(password, str):
            password = password.encode('utf-8')

        m = hashlib.sha256()
        m.update(password)
        hash_pw = m.digest()

        m = hashlib.sha256()
        m.update(password)
        m.update(hash_pw)
        dec_key = m.digest()

        try:
            nonce = hash_pw[4:16]
            aesgcm = AESGCM(dec_key)
            dec_value = aesgcm.decrypt(nonce=nonce,
                                       data=encrypted_bytes,
                                       associated_data=b'')
        except InvalidTag as e:
            raise GeneralException("Fail to decrypt an account. Please check the password.") from e

        return Account(private_key=dec_value)

    def verify_inclusion(self, root):
        """ verify_inclusion verifies the contract state is included in the
        general trie root.
        """
        if self.__state_proof is None:
            return False
        if isinstance(root, str) and len(root) != 0:
            root = decode_root(root)
        key = hashlib.sha256(self.__address).digest()
        value = hashlib.sha256(self.__state.SerializeToString()).digest()
        ap = self.__state_proof.auditPath
        if self.__state_proof.bitmap:
            height = self.__state_proof.height
            bitmap = self.__state_proof.bitmap
            return mp.verify_inclusion_c(ap, height, bitmap, root, key, value)
        return mp.verify_inclusion(ap, root, key, value)

    def verify_exclusion(self, root):
        """ verify_exclusion verifies that the contract state doesnt exist
        in the general trie root.
        """
        if self.__state_proof is None:
            return False
        key = hashlib.sha256(self.__address).digest()
        if self.__state_proof.bitmap:
            return mp.verify_exclusion_c(root,
                                         self.__state_proof.auditPath,
                                         self.__state_proof.height,
                                         self.__state_proof.bitmap,
                                         key,
                                         self.__state_proof.proofKey,
                                         self.__state_proof.proofVal)
        return mp.verify_exclusion(root,
                                   self.__state_proof.auditPath,
                                   key,
                                   self.__state_proof.proofKey,
                                   self.__state_proof.proofVal)



import sys
import traceback
import json

import aergo.herapy as herapy
from aergo.herapy.utils.converter import convert_bytes_to_hex_str, \
                                         convert_bytes_to_int_str


def run():
    try:
        aergo1 = herapy.Aergo()
        aergo2 = herapy.Aergo()
        aergo3 = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo1.connect('localhost:17801')
        aergo2.connect('localhost:17802')
        aergo3.connect('localhost:17803')

        print("------ Get Blockchain Status -----------")
        best_block_hash, best_block_height = aergo1.get_blockchain_status()
        print("(aergo1) Best Block Hash      = {}".format(best_block_hash))
        print("(aergo1) Best Block Height    = {}".format(best_block_height))

        best_block_hash, best_block_height = aergo2.get_blockchain_status()
        print("(aergo2) Best Block Hash      = {}".format(best_block_hash))
        print("(aergo2) Best Block Height    = {}".format(best_block_height))

        best_block_hash, best_block_height = aergo3.get_blockchain_status()
        print("(aergo3) Best Block Hash      = {}".format(best_block_hash))
        print("(aergo3) Best Block Height    = {}".format(best_block_height))

        print("------ Get Block Status -----------")
        block = aergo1.get_block(block_height=best_block_height)
        print("Aergo 1:\n{}".format(json.dumps(block.json(), indent=2)))

        aergo1.disconnect()
        aergo2.disconnect()
        aergo3.disconnect()
    except Exception:
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



import pytest

import ecdsa
import hashlib

from aergo.herapy.obj.address import Address


def test_fail():
    with pytest.raises(AssertionError):
        Address(None)

    with pytest.raises(AttributeError):
        Address(1234)

    with pytest.raises(ValueError):
        Address("1234")


def test_success():
    sk = ecdsa.SigningKey.generate(curve=ecdsa.SECP256k1,
                                   hashfunc=hashlib.sha256)
    addr = Address(sk.privkey.public_key)
    assert type(addr) is Address
    assert isinstance(str(addr), str)
    assert isinstance(bytes(addr), bytes)



import sys
import traceback

import aergo.herapy as herapy


def run():
    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Get Accounts without State in Node -----------")
        accounts = aergo.get_node_accounts(skip_state=True)
        for i, account in enumerate(accounts):
            print("  > account state : {}".format(account))
            print("    - balance")
            print("        + aergo      = {}".format(account.balance.aergo))
            print("        + gaer       = {}".format(account.balance.gaer))
            print("        + aer        = {}".format(account.balance.aer))
            print("    - nonce          = {}".format(account.nonce))
            print("    - code hash      = {}".format(account.code_hash))
            print("    - storage root   = {}".format(account.storage_root))

        print("------ Get Accounts in Node -----------")
        accounts = aergo.get_node_accounts()
        for i, account in enumerate(accounts):
            print("  > account state : {}".format(account))
            print("    - balance")
            print("        + aergo      = {}".format(account.balance.aergo))
            print("        + gaer       = {}".format(account.balance.gaer))
            print("        + aer        = {}".format(account.balance.aer))
            print("    - nonce          = {}".format(account.nonce))
            print("    - code hash      = {}".format(account.code_hash))
            print("    - storage root   = {}".format(account.storage_root))

        print("------ Create Account -----------")
        password = "test password"
        account = aergo.new_account()

        print("Private Key      = {}".format(account.private_key))
        print("Public Key       = {}".format(account.public_key))
        print("Address          = {}".format(account.address))

        print("------ Get Account State -----------")
        aergo.get_account()
        print("  > account state in 'aergo'")
        print('    - Nonce:        %s' % aergo.account.nonce)
        print("    - balance")
        print("        + aergo      = {}".format(aergo.account.balance.aergo))
        print("        + gaer       = {}".format(aergo.account.balance.gaer))
        print("        + aer        = {}".format(aergo.account.balance.aer))
        print('    - Code Hash:    %s' % aergo.account.code_hash)
        print('    - Storage Root: %s' % aergo.account.storage_root)
        print(account)

        print("------ Get Configured Account -----------")
        conf_keys_str = """
1:eHoEcHnaxpGpgzknXjuwon8VFVrLkKHC4FckGuGkQ8depiDDfyUAWC3L:AmPZKCJpT98V9Tc8dBUbRg78M1jgoB1ZEh97Rs1r5KewPcCiURf7
"""
        conf_keys_line = conf_keys_str.split('\n')
        conf_keys = []
        for l in conf_keys_line:
            if 0 == len(l):
                continue
            conf_keys.append(l.split(':'))
        print("  - Configured Keys:\n{}".format(conf_keys))

        for k in conf_keys:
            print("  [{}]".format(k[0]))
            print("    > private key   : {}".format(k[1]))
            print("    > address       : {}".format(k[2]))

            # check account state
            a = aergo.get_account(address=k[2])
            print("    > account state : {}".format(a))
            print("      - balance        = {}".format(a.balance))
            print("        + aergo          = {}".format(aergo.account.balance.aergo))
            print("        + gaer           = {}".format(aergo.account.balance.gaer))
            print("        + aer            = {}".format(aergo.account.balance.aer))
            print("      - nonce          = {}".format(a.nonce))
            print("      - code hash      = {}".format(a.code_hash))
            print("      - storage root   = {}".format(a.storage_root))

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception:
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: node.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='node.proto',
  package='types',
  syntax='proto3',
  serialized_options=_b('Z\036github.com/aergoio/aergo/types'),
  serialized_pb=_b('\n\nnode.proto\x12\x05types\"<\n\x0bPeerAddress\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\r\x12\x0e\n\x06peerID\x18\x03 \x01(\x0c\x42 Z\x1egithub.com/aergoio/aergo/typesb\x06proto3')
)




_PEERADDRESS = _descriptor.Descriptor(
  name='PeerAddress',
  full_name='types.PeerAddress',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='address', full_name='types.PeerAddress.address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port', full_name='types.PeerAddress.port', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='peerID', full_name='types.PeerAddress.peerID', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21,
  serialized_end=81,
)

DESCRIPTOR.message_types_by_name['PeerAddress'] = _PEERADDRESS
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PeerAddress = _reflection.GeneratedProtocolMessageType('PeerAddress', (_message.Message,), dict(
  DESCRIPTOR = _PEERADDRESS,
  __module__ = 'node_pb2'
  # @@protoc_insertion_point(class_scope:types.PeerAddress)
  ))
_sym_db.RegisterMessage(PeerAddress)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

from ecdsa.ecdsa import int_to_string

from ..utils.encoding import encode_address, decode_address


class Address:
    def __init__(self, pubkey):
        if pubkey is None:
            assert 1 == 0

        if isinstance(pubkey, str):
            pubkey = decode_address(pubkey)

        if isinstance(pubkey, bytes):
            self.__address = pubkey
        else:
            self.__generate_address(pubkey)

    def __str__(self):
        return encode_address(self.__address)

    def __bytes__(self):
        return self.__address

    def __generate_address(self, pubkey):
        pubkey_x = pubkey.point.x()
        x_bytes = int_to_string(pubkey_x)
        pubkey_y = pubkey.point.y()
        head = bytes([2] if 0 == pubkey_y % 2 else [3])
        self.__address = head + x_bytes



# -*- coding: utf-8 -*-

"""Main module."""

import json

from . import account as acc
from . import comm

from .obj import block
from .obj import transaction
from .obj import address as addr
from .obj import block_hash as bh
from .obj import peer as pr
from .obj import tx_hash as th
from .obj.call_info import CallInfo
from .obj.tx_result import TxResult
from .obj.sc_state import SCState
from .obj.var_proof import VarProof

from .errors.exception import CommunicationException
from .errors.general_exception import GeneralException

from .status.commit_status import CommitStatus
from .status.smartcontract_status import SmartcontractStatus

from .utils.encoding import encode_address, decode_address, \
    encode_private_key, decode_private_key, decode_root, \
    encode_tx_hash, decode_tx_hash


class Aergo:
    def __init__(self):
        self.__account = None
        self.__comm = None

    @property
    def account(self):
        """
        Returns the account object.
        :return:
        """
        return self.__account

    @account.setter
    def account(self, a):
        self.__account = a

    def new_account(self, private_key=None, skip_state=False):
        self.__account = acc.Account(private_key=private_key)
        if not skip_state:
            self.get_account()
        return self.__account

    # TODO how about making account_state class,
    #       or how about returning account and change method name
    def get_account(self, account=None, address=None, proof=False, root=b'', compressed=True):
        """
        Return account information
        :param address:
        :param proof:
        :param root:
        :param compressed:
        :return:
        """
        if self.__comm is None:
            return None

        if account is None and address is None:
            # self account
            ret_account = self.__account
            req_address = bytes(self.__account.address)
        elif account is not None:
            ret_account = account
            req_address = bytes(account.address)
        else:
            if address is None:
                raise GeneralException("Fail to get an account info. No designated account address.")
            else:
                ret_account = acc.Account(empty=True)
                if isinstance(address, str):
                    req_address = decode_address(address)
                elif type(address) is addr.Address:
                    req_address = bytes(address)
                else:
                    req_address = address
                ret_account.address = req_address

        if proof:
            if isinstance(root, str) and len(root) != 0:
                root = decode_root(root)
            try:
                state_proof = self.__comm.get_account_state_proof(req_address,
                                                                  root,
                                                                  compressed)
            except Exception as e:
                raise CommunicationException(e) from e

            ret_account.state = state_proof.state
            ret_account.state_proof = state_proof
        else:
            try:
                state = self.__comm.get_account_state(req_address)
            except Exception as e:
                raise CommunicationException(e) from e

            ret_account.state = state
            ret_account.state_proof = None

        return ret_account

    def connect(self, target):
        """
        Connect to the gRPC server running on port `target` e.g.
        target="localhost:7845".
        :param target:
        :return:
        """
        if target is None:
            raise ValueError('need target value')

        self.__comm = comm.Comm(target)
        try:
            self.__comm.connect()
        except Exception as e:
            raise CommunicationException(e) from e

    def disconnect(self):
        """
        Disconnect from the gRPC server.
        """
        if self.__comm is not None:
            try:
                self.__comm.disconnect()
            except Exception as e:
                raise CommunicationException(e) from e

    def get_blockchain_status(self):
        """
        Returns the highest block hash and block height so far.
        :return:
        """
        if self.__comm is None:
            return None, -1

        try:
            status = self.__comm.get_blockchain_status()
        except Exception as e:
            raise CommunicationException(e) from e

        return bh.BlockHash(status.best_block_hash), status.best_height

    def get_block(self, block_hash=None, block_height=-1):
        """
        Returns information about block `block_hash`.
        :param block_hash:
        :param block_height:
        :return:
        """
        if self.__comm is None:
            return None

        if block_height > 0:
            query = block_height.to_bytes(8, byteorder='little')
        else:
            if type(block_hash) is not bh.BlockHash:
                block_hash = bh.BlockHash(block_hash)
            query = block_hash.value

        try:
            result = self.__comm.get_block(query)
        except Exception as e:
            raise CommunicationException(e) from e

        b = block.Block(grpc_block=result)
        return b

    def get_node_accounts(self, skip_state=False):
        """
        Returns a list of all node accounts.
        :return:
        """

        try:
            result = self.__comm.get_accounts()
        except Exception as e:
            raise CommunicationException(e) from e

        accounts = []
        for a in result.accounts:
            if skip_state:
                account = acc.Account(empty=True)
                account.address = a.address
            else:
                account = self.get_account(address=a.address)
            accounts.append(account)

        return accounts

    def get_peers(self):
        """
        Returns a list of peers.
        :return:
        """
        try:
            result = self.__comm.get_peers()
        except Exception as e:
            raise CommunicationException(e) from e

        peers = []
        for i, p in enumerate(result.peers):
            peer = pr.Peer()
            peer.info = p
            peers.append(peer)

        return peers

    def get_node_state(self, timeout=1):
        """
        Returns information about the node state.
        :return:
        """
        try:
            result = self.__comm.get_node_state(timeout)
        except Exception as e:
            raise CommunicationException(e) from e

        json_txt = result.value.decode('utf8').replace("'", '"')
        return json.loads(json_txt)

    def get_tx(self, tx_hash):
        """
        Returns info on transaction with hash `tx_hash`.
        :param tx_hash:
        :return:
        """
        try:
            result = self.__comm.get_tx(tx_hash)
        except Exception as e:
            raise CommunicationException(e) from e

        return result

    def lock_account(self, address, passphrase):
        """
        Locks the account with address `address` with the passphrase
        `passphrase`.
        :param address:
        :param passphrase:
        :return:
        """
        try:
            result = self.__comm.lock_account(address, passphrase)
        except Exception as e:
            raise CommunicationException(e) from e

        return result

    def unlock_account(self, address, passphrase):
        """
        Unlocks the account with address `address` with the passphrase
        `passphrase`.
        :param address:
        :param passphrase:
        :return:
        """
        try:
            result = self.__comm.unlock_account(address=address,
                                                passphrase=passphrase)
        except Exception as e:
            raise CommunicationException(e) from e

        return result

    @staticmethod
    def _generate_tx(account, to_address, nonce, amount, fee_limit, fee_price,
                     payload):
        if to_address is not None:
            to_address = addr.Address(to_address)

        tx = transaction.Transaction(from_address=account.address,
                                     to_address=to_address,
                                     nonce=nonce, amount=amount,
                                     fee_limit=fee_limit, fee_price=fee_price,
                                     payload=payload)
        tx.sign = account.sign_msg_hash(tx.calculate_hash(including_sign=False))
        return tx

    def send_payload(self, amount, payload, to_address=None, retry_nonce=0):
        if self.__comm is None:
            return None, None

        if isinstance(to_address, str):
            to_address = decode_address(to_address)

        nonce = self.__account.nonce + 1
        tx = self._generate_tx(account=self.__account, to_address=to_address,
                               nonce=nonce, amount=amount,
                               fee_limit=0, fee_price=0,
                               payload=payload)
        signed_txs, results = self.send_tx(tx)

        if results[0].status == CommitStatus.TX_OK:
            self.__account.nonce = nonce
        elif results[0].status == CommitStatus.TX_HAS_SAME_NONCE:
            while retry_nonce > 0:
                retry_nonce -= 1

                nonce += 1
                tx = self._generate_tx(account=self.__account,
                                       to_address=to_address,
                                       nonce=nonce, amount=amount,
                                       fee_limit=0, fee_price=0,
                                       payload=payload)
                signed_txs, results = self.send_tx(tx)

                es = results[0].status
                if es == CommitStatus.TX_OK:
                    self.__account.nonce = nonce
                    break
                elif es != CommitStatus.TX_HAS_SAME_NONCE:
                    break

        return signed_txs[0], results[0]

    def send_unsigned_tx(self, unsigned_tx):
        """
        Sends the unsigned transaction.
        The unsigned transaction will be signed by the account
        which is stored in the connected node.
        :param unsigned_tx:
        :return:
        """
        try:
            result = self.__comm.send_tx(unsigned_tx)
        except Exception as e:
            raise CommunicationException(e) from e

        return result

    def send_tx(self, signed_txs):
        """
        Send a set of signed transactions simultaneously.
        These transactions will push to the memory pool after verifying.
        :param signed_txs:
        :return:
        """
        if not isinstance(signed_txs, (list, tuple)):
            signed_txs = [signed_txs]

        try:
            result_list = self.__comm.commit_txs(signed_txs)
        except Exception as e:
            raise CommunicationException(e) from e

        results = []
        for i, r in enumerate(result_list.results):
            tx_result = TxResult(signed_txs[i], result_list.results[i])
            results.append(tx_result)
        return signed_txs, results

    def import_account(self, exported_data, password, skip_state=False, skip_self=False):
        if exported_data is None or 0 == len(exported_data):
            # TODO unit test + exception handling
            assert 1 == 0

        if password is None or 0 == len(password):
            # TODO unit test + exception handling
            assert 1 == 0

        if isinstance(exported_data, str):
            exported_data = decode_private_key(exported_data)

        if isinstance(password, bytes):
            password = password.decode('utf-8')

        account = acc.Account.decrypt_account(exported_data, password)
        if not skip_self:
            self.__account = account
            if not skip_state:
                self.get_account()
        else:
            if not skip_state:
                self.get_account(account=account)
        return account

    def export_account(self, password, account=None):
        if account is None:
            account = self.__account

        enc_acc = acc.Account.encrypt_account(account, password)
        return encode_private_key(enc_acc)

    def get_tx_result(self, tx_hash):
        if self.__comm is None:
            return None

        if isinstance(tx_hash, str):
            tx_hash = decode_tx_hash(tx_hash)
        elif type(tx_hash) is th.TxHash:
            tx_hash = bytes(tx_hash)

        try:
            result = self.__comm.get_receipt(tx_hash)
            tx_result = TxResult()
            tx_result.tx_id = encode_tx_hash(tx_hash)
            try:
                tx_result.status = SmartcontractStatus(result.status)
                tx_result.detail = result.ret

                if 'error' in result.ret:
                    raise ValueError
            except ValueError:
                tx_result.status = SmartcontractStatus.ERROR
                if 'CREATED' == result.status:
                    tx_result.detail = result.ret
                else:
                    tx_result.detail = result.status
            tx_result.contract_address = encode_address(result.contractAddress)
        except Exception as e:
            raise CommunicationException(e) from e

        return tx_result

    def deploy_sc(self, payload, amount=0, args=None):
        if isinstance(payload, str):
            payload = decode_address(payload)

        if args is not None and not isinstance(args, (list, tuple)):
            args = [args]

        payload_bytes = (len(payload) + 4).to_bytes(4, byteorder='little')
        payload_bytes += payload

        json_args = json.dumps(args, separators=(',', ':'))
        payload_bytes += json_args.encode('utf-8')

        tx, result = self.send_payload(amount=amount, payload=payload_bytes)
        return tx, result

    def new_call_sc_tx(self, sc_address, func_name, amount=0, args=None):
        if isinstance(sc_address, str):
            # TODO exception handling: raise ValueError("Invalid checksum")
            sc_address = decode_address(sc_address)

        if args is not None and not isinstance(args, (list, tuple)):
            args = [args]

        call_info = CallInfo(func_name, args).__dict__
        payload = json.dumps(call_info, separators=(',', ':')).encode('utf-8')

        self.__account.nonce += 1
        return self._generate_tx(account=self.__account, to_address=sc_address,
                                 nonce=self.__account.nonce, amount=amount,
                                 fee_limit=0, fee_price=0, payload=payload)

    def batch_call_sc(self, sc_txs):
        return self.send_tx(sc_txs)

    def call_sc(self, sc_address, func_name, amount=0, args=None):
        sc_tx = self.new_call_sc_tx(sc_address=sc_address, func_name=func_name,
                                    amount=amount, args=args)
        sc_txs, results = self.batch_call_sc([sc_tx])
        return sc_txs[0], results[0]

    def query_sc(self, sc_address, func_name, args=None):
        if isinstance(sc_address, str):
            # TODO exception handling: raise ValueError("Invalid checksum")
            sc_address = decode_address(sc_address)

        if args is not None and not isinstance(args, (list, tuple)):
            args = [args]

        call_info = CallInfo(func_name, args).__dict__
        payload = json.dumps(call_info, separators=(',', ':')).encode('utf-8')

        try:
            result = self.__comm.query_contract(sc_address, payload)
        except Exception as e:
            raise CommunicationException(e) from e

        return result.value

    def query_sc_state(self, sc_address, var_name, var_index="", root=b'',
                       compressed=True):
        """ query_sc_state returns a SCState object containing the contract
        state and variable state with their respective merkle proofs.
        """
        if isinstance(sc_address, str):
            # TODO exception handling: raise ValueError("Invalid checksum")
            sc_address = decode_address(sc_address)
        if isinstance(root, str) and len(root) != 0:
            root = decode_root(root)
        try:
            result = self.__comm.query_contract_state(sc_address, var_name,
                                                      var_index, root,
                                                      compressed)
        except Exception as e:
            raise CommunicationException(e) from e
        var_proof = VarProof(result.varProof, var_name, var_index)
        account = acc.Account(empty=True)
        account.state = result.contractProof.state
        account.state_proof = result.contractProof
        account.address = sc_address
        return SCState(account, var_proof)



# -*- coding: utf-8 -*-

import hashlib

from google.protobuf.json_format import MessageToJson
from aergo.herapy.utils import merkle_proof as mp


class VarProof:
    """ VarProof holds the inclusion/exclusion proof of a variable state
    inside a contract state trie
    """
    def __init__(self, var_proof, var_name, var_index=""):
        self.__var_proof = var_proof
        self.__var_name = var_name
        self.__var_index = var_index
        if len(var_index) != 0:
            var_id = bytes("_sv_" + var_name + "-" + var_index, "utf-8")
        else:
            var_id = bytes("_sv_" + var_name, "utf-8")
        self.__trie_key = hashlib.sha256(var_id).digest()

    def __str__(self):
        return MessageToJson(self.__var_proof)

    @property
    def var_index(self):
        return self.__var_index

    @property
    def var_name(self):
        return self.__var_name

    @property
    def trie_key(self):
        return self.__trie_key

    @property
    def var_proof(self):
        return self.__var_proof

    @var_proof.setter
    def var_proof(self, v):
        self.__var_proof = v

    def verify_inclusion(self, root):
        """ verify_inclusion verifies the contract state is included in the
        contract trie root.
        """
        if self.__var_proof is None:
            return False
        key = self.__trie_key
        value = hashlib.sha256(self.__var_proof.value).digest()
        ap = self.__var_proof.auditPath
        # if bitmap exists, then the merkle proof is compressed
        if self.__var_proof.bitmap:
            height = self.__var_proof.height
            bitmap = self.__var_proof.bitmap
            return mp.verify_inclusion_c(ap, height, bitmap, root, key, value)
        return mp.verify_inclusion(ap, root, key, value)

    def verify_exclusion(self, root):
        """ verify_exclusion verifies that the contract variable state
        doesnt exist in the contract trie root.
        """
        if self.__var_proof is None:
            return False
        # if bitmap exists, then the merkle proof is compressed
        if self.__var_proof.bitmap:
            return mp.verify_exclusion_c(root,
                                         self.__var_proof.auditPath,
                                         self.__var_proof.height,
                                         self.__var_proof.bitmap,
                                         self.__trie_key,
                                         self.__var_proof.proofKey,
                                         self.__var_proof.proofVal)
        return mp.verify_exclusion(root,
                                   self.__var_proof.auditPath,
                                   self.__trie_key,
                                   self.__var_proof.proofKey,
                                   self.__var_proof.proofVal)



# -*- coding: utf-8 -*-

from google.protobuf.json_format import MessageToJson


class SCState:
    """ SCState holds the inclusion/exclusion proofs of a contract
    state in the global trie and of a variable's value in the
    contract trie.
    SCState is returned by aergo.query_sc_state() for easy merkle
    proof verification give a root.
    """
    def __init__(self, account, var_proof):
        self.__account = account
        self.__var_proof = var_proof

    def __str__(self):
        account_str = MessageToJson(self.__account.state_proof)
        var_str = str(self.__var_proof)
        return account_str + var_str

    @property
    def account(self):
        return self.__account

    @property
    def var_proof(self):
        return self.__var_proof

    @var_proof.setter
    def var_proof(self, v):
        self.__var_proof = v

    def verify_inclusion(self, root):
        """ verify_inclusion verifies the contract state is included in the
        general trie root and the variable state is included in that contract
        state.
        """
        if not self.__account.verify_inclusion(root):
            # The contract state doesnt exist
            return False
        sc_root = self.__account.state_proof.state.storageRoot
        # Verify the variable state is included in the contract root
        return self.__var_proof.verify_inclusion(sc_root)

    def verify_exclusion(self, root):
        """ verify_exclusion verifies that the contract state doesnt exist,
        if it does then verify the variable is not included in that
        contract state.
        """
        if self.__account.verify_exclusion(root):
            # The contract state doesnt exist
            return True
        # First verify that the contract state does exist before proving
        # that the variable state doesnt exist
        if not self.__account.verify_inclusion(root):
            # The contract state doesnt exist so the variable exclusion proof
            # cannot be valid
            return False
        sc_root = self.__account.state_proof.state.storageRoot
        return self.__var_proof.verify_exclusion(sc_root)
